/*
 * abydos loader template
 *
 * Copyright (C) 2004 - 2018 Magnus Bergman
 *
 * Authors: Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <abydos.h>
#include <cairo.h>
#include <math.h>
#include <gdk-pixbuf/gdk-pixbuf-io.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>                     /* memcpy() */

#ifndef MODULE_ENTRY
# define MODULE_ENTRY(a,b) b
#endif
#define seconds_from_timeval(tv) ((tv)->tv_sec+(double)(tv)->tv_usec/1000000)


static void
blit_surface_pixbuf(cairo_surface_t *surface,GdkPixbuf *pixbuf,cairo_rectangle_int_t *r)
{
    typedef struct{
        uint8_t r;
        uint8_t g;
        uint8_t b;
        uint8_t a;
    }pixel_t;
    int src_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    const uint32_t *src = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = gdk_pixbuf_get_rowstride(pixbuf)/sizeof(pixel_t);
    pixel_t *dst = (pixel_t*)gdk_pixbuf_get_pixels(pixbuf);
    int x;
    int y;

    dst += r->x + r->y * dst_stride;
    for(y=0;y<r->height;++y){
        for(x=0;x<r->width;++x){
            /* FIXME!! undo premultiply */
            dst->r = (*src >> 16) & 0xff;
            dst->g = (*src >>  8) & 0xff;
            dst->b = (*src >>  0) & 0xff;
            dst->a = (*src >> 24) & 0xff;
            ++src;
            ++dst;
        }
        src += src_stride - r->width;
        dst += dst_stride - r->width;
    }
}


static void
blit_abydos_pixbuf(abydos_t *ar,GdkPixbuf *pixbuf,cairo_rectangle_int_t *r)
{
    cairo_surface_t *surface;
    cairo_t *cr;
    if(!pixbuf)
        return;
    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,r->width,r->height);
    cr = cairo_create(surface);
    cairo_translate(cr,-r->x,-r->y);
    abydos_render(ar,cr);
    cairo_destroy(cr);
    blit_surface_pixbuf(surface,pixbuf,r);
    cairo_surface_destroy(surface);
}


static GdkPixbuf *
abydos_create_pixbuf(abydos_t *ar)
{
    cairo_rectangle_int_t r;
    cairo_surface_t *surface = abydos_get_image_surface(ar);
    GdkPixbuf *pixbuf;
    r.x = 0;
    r.y = 0;
    abydos_get_size(ar,&r.width,&r.height);
    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,TRUE,8,r.width,r.height);
    blit_surface_pixbuf(surface,pixbuf,&r);
    cairo_surface_destroy(surface);
    return pixbuf;
}


#ifdef CAN_ANIMATE
/* --- animation -------------------------------------------------------- */

typedef struct {
    GdkPixbufAnimationClass inh;
} AbydosAnimClass;

typedef struct {
    GdkPixbufAnimation inh;
    abydos_t *ar;
} AbydosAnim;

G_DEFINE_TYPE (AbydosAnim, abydos_anim, GDK_TYPE_PIXBUF_ANIMATION);

typedef struct {
    GdkPixbufAnimationIterClass inh;
} AbydosAnimIterClass;

typedef struct {
    GdkPixbufAnimationIter inh;
    abydos_t *ar;
    double start_time;
    //double current_time;
} AbydosAnimIter;

G_DEFINE_TYPE (AbydosAnimIter, abydos_anim_iter, GDK_TYPE_PIXBUF_ANIMATION_ITER);


static void
abydos_anim_init (AbydosAnim *anim)
{
}


static void
abydos_anim_finalize (GObject *object)
{
    AbydosAnim *anim = (AbydosAnim*)object;
    abydos_destroy(anim->ar);
    G_OBJECT_CLASS (abydos_anim_parent_class)->finalize (object);
}


static gboolean
abydos_anim_is_static_image  (GdkPixbufAnimation *animation)
{
    AbydosAnim *anim = (AbydosAnim*)animation;
    return abydos_get_frame_count(anim->ar) <= 1;
}


static GdkPixbuf*
abydos_anim_get_static_image (GdkPixbufAnimation *animation)
{
    AbydosAnim *anim = (AbydosAnim*)animation;
    return abydos_create_pixbuf(anim->ar);
}


static void
abydos_anim_get_size (GdkPixbufAnimation *animation,
                      int                *width,
                      int                *height)
{
    AbydosAnim *anim = (AbydosAnim*)animation;
    abydos_get_size(anim->ar,width,height);
}


static GdkPixbufAnimationIter*
abydos_anim_get_iter (GdkPixbufAnimation *animation,
                              const GTimeVal     *start_time)
{
    AbydosAnim *anim = (AbydosAnim*)animation;
    AbydosAnimIter *iter = g_object_new (abydos_anim_iter_get_type(), NULL);
    iter->ar = abydos_reference(anim->ar);

    iter->start_time = seconds_from_timeval(start_time);
    //iter->first_loop_slowness = 0;

    return GDK_PIXBUF_ANIMATION_ITER (iter);
}


static void
abydos_anim_class_init (AbydosAnimClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GdkPixbufAnimationClass *anim_class = GDK_PIXBUF_ANIMATION_CLASS (klass);

    object_class->finalize = abydos_anim_finalize;

    anim_class->is_static_image = abydos_anim_is_static_image;
    anim_class->get_static_image = abydos_anim_get_static_image;
    anim_class->get_size = abydos_anim_get_size;
    anim_class->get_iter = abydos_anim_get_iter;
}


static void
abydos_anim_iter_init (AbydosAnimIter *iter)
{
}


static void
abydos_anim_iter_finalize (GObject *object)
{
    AbydosAnimIter *iter = (AbydosAnimIter*)object;
    abydos_destroy(iter->ar);
    G_OBJECT_CLASS (abydos_anim_iter_parent_class)->finalize (object);
}


static gboolean
abydos_anim_iter_advance (GdkPixbufAnimationIter *anim_iter,
                          const GTimeVal         *tv_now)
{
    AbydosAnimIter *iter = (AbydosAnimIter*)anim_iter;
    double now = seconds_from_timeval(tv_now);

    if (now < iter->start_time) {
        /* Try to compensate; probably the system clock
         * was set backwards
         */
        iter->start_time = now;
    }
    return abydos_set_time(iter->ar,now - iter->start_time);
}


static int
abydos_anim_iter_get_delay_time (GdkPixbufAnimationIter *anim_iter)
{
    AbydosAnimIter *iter = (AbydosAnimIter*)anim_iter;
    return ceil(abydos_get_delay(iter->ar)*1000);
}


static GdkPixbuf*
abydos_anim_iter_get_pixbuf (GdkPixbufAnimationIter *anim_iter)
{
    AbydosAnimIter *iter = (AbydosAnimIter*)anim_iter;
    return abydos_create_pixbuf(iter->ar);
}


static gboolean
abydos_anim_iter_on_currently_loading_frame (GdkPixbufAnimationIter *anim_iter)
{
    /* not supported (yet) */
    return FALSE;
}


static void
abydos_anim_iter_class_init (AbydosAnimIterClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GdkPixbufAnimationIterClass *anim_iter_class = GDK_PIXBUF_ANIMATION_ITER_CLASS (klass);

    object_class->finalize = abydos_anim_iter_finalize;

    anim_iter_class->get_delay_time = abydos_anim_iter_get_delay_time;
    anim_iter_class->get_pixbuf = abydos_anim_iter_get_pixbuf;
    anim_iter_class->on_currently_loading_frame = abydos_anim_iter_on_currently_loading_frame;
    anim_iter_class->advance = abydos_anim_iter_advance;
}


static AbydosAnim *
abydos_anim_new(abydos_t *ar)
{
    AbydosAnim *anim = g_object_new(abydos_anim_get_type(),NULL);
    anim->ar = ar;
    return anim;
}
#endif /* CAN_ANIMATE */

/* ---------------------------------------------------------------------- */

typedef struct _Context Context;
struct _Context{
    abydos_t *ar;
    GdkPixbufModuleSizeFunc size_func;
    GdkPixbufModulePreparedFunc prepared_func;
    GdkPixbufModuleUpdatedFunc updated_func;
    void *userdata;
    int width;
    int height;
    GdkPixbuf *pixbuf;
#ifdef CAN_ANIMATE
    GdkPixbufAnimation *animation;
#endif
};


static void
on_info(void *userdata,abydos_t *ar)
{
    Context *c = (Context*)userdata;
    abydos_info_t info;

    abydos_get_info(c->ar,&info);
    c->width = info.width;
    c->height = info.height;
    if(c->size_func)
        c->size_func(&c->width,&c->height,c->userdata);
    if(c->width > 0 && c->height > 0){
        c->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,TRUE,8,c->width,c->height);
#ifdef CAN_ANIMATE
        c->animation = (GdkPixbufAnimation*)abydos_anim_new(abydos_reference(ar));
#endif
        if(c->prepared_func){
#ifdef CAN_ANIMATE
            c->prepared_func(c->pixbuf,c->animation,c->userdata);
#else
            c->prepared_func(c->pixbuf,NULL,c->userdata);
#endif
        }
    }else{
        c->pixbuf = NULL;
#ifdef CAN_ANIMATE
        c->animation = NULL;
#endif
    }
}


static void
on_update(void *userdata,abydos_t *ar,cairo_rectangle_int_t *r)
{
    Context *c = (Context*)userdata;
    blit_abydos_pixbuf(ar,c->pixbuf,r);
    if(c->updated_func && c->pixbuf)
        c->updated_func(c->pixbuf,r->x,r->y,r->width,r->height,c->userdata);
}


static void *
begin_load(GdkPixbufModuleSizeFunc size_func,
           GdkPixbufModulePreparedFunc prepared_func,
           GdkPixbufModuleUpdatedFunc updated_func,
           void *userdata,
           GError** err)
{
    Context *c = g_new(Context,1);
    c->ar = abydos_create(MIME_TYPE);
    if(!c->ar){
        g_set_error(err,
                    GDK_PIXBUF_ERROR,
                    GDK_PIXBUF_ERROR_UNKNOWN_TYPE,
                    "No abydos plugin found for " MIME_TYPE);
        g_free(c);
        return NULL;
    }
    c->size_func = size_func;
    c->prepared_func = prepared_func;
    c->updated_func = updated_func;
    c->userdata = userdata;
    abydos_load_begin(c->ar,on_info,on_update,c);

    return c;
}


static gboolean
load_increment(void *handle,
               const unsigned char *data,
               unsigned int len,
               GError** err)
{
    Context *c = (Context*)handle;
    if(abydos_load_feed(c->ar,(char*)data,len)<0){
        g_set_error_literal(err,
                            GDK_PIXBUF_ERROR,
                            GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                            abydos_error(c->ar));
        return FALSE;
    }
    return TRUE;
}


static gboolean stop_load(void *handle,GError** err){
    Context *c = (Context*)handle;
    int res = abydos_load_end(c->ar);
    if(res<0){
        g_set_error_literal(err,
                            GDK_PIXBUF_ERROR,
                            GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                            abydos_error(c->ar));
    }
    abydos_destroy(c->ar);
    g_free(c);
    return 0==res;
}


static GdkPixbuf* load(FILE* f,GError** err)
{
    GdkPixbuf* pixbuf = NULL;
    abydos_t *ar = abydos_create(MIME_TYPE);
    if(!ar){
        g_set_error(err,
                    GDK_PIXBUF_ERROR,
                    GDK_PIXBUF_ERROR_UNKNOWN_TYPE,
                    "No abydos plugin found for " MIME_TYPE);
        return NULL;
    }
    abydos_load_begin(ar,NULL,NULL,NULL);
    for(;;){
        char buffer[4096];
        ssize_t bytes = fread(buffer,1,4096,f);
        if(bytes<0)
            break;
        if(abydos_load_feed(ar,buffer,bytes)<0){
            g_set_error_literal(err,
                                GDK_PIXBUF_ERROR,
                                GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                                abydos_error(ar));
            goto out;
        }
    }
    if(abydos_load_end(ar)<0){
        g_set_error_literal(err,
                            GDK_PIXBUF_ERROR,
                            GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                            abydos_error(ar));
        goto out;
    }
    pixbuf = abydos_create_pixbuf(ar);
out:
    abydos_destroy(ar);
    return pixbuf;
}


void MODULE_ENTRY (PREFIX, fill_vtable) (GdkPixbufModule* module){
    module->begin_load     = begin_load;
    module->stop_load      = stop_load;
    module->load_increment = load_increment;
    module->load = load;
}


void MODULE_ENTRY (PREFIX, fill_info) (GdkPixbufFormat* format){
    static GdkPixbufModulePattern pattern[] = {
PATTERNS
        {NULL,NULL}
    };
    static char* mime_types[] = {
MIME_TYPES
        NULL
    };
    static char* extensions[] = {
EXTENSIONS
        NULL
    };

    format->name = NAME;
    format->signature = pattern;
    format->domain = NULL;
    format->description = DESCRIPTION;
    format->mime_types = mime_types;
    format->extensions = extensions;
    format->flags = 0;
    format->disabled = FALSE;
    format->license = "LGPL";
}
